@@ -1,4 +1,4 @@
-{-# LANGUAGE FlexibleContexts, FlexibleInstances, FunctionalDependencies, GeneralizedNewtypeDeriving, MultiParamTypeClasses, RankNTypes, TypeFamilies, UndecidableInstances  #-}
+{-# LANGUAGE CPP, FlexibleContexts, FlexibleInstances, FunctionalDependencies, GeneralizedNewtypeDeriving, MultiParamTypeClasses, RankNTypes, TypeFamilies, UndecidableInstances  #-}
 {-| This module defines the Monad stack used by Happstack. You mostly don't want to be looking in here. Look in "Happstack.Server.Monads" instead.
 -}
 module Happstack.Server.Internal.Monads where
@@ -8,6 +8,7 @@ import Control.Applicative                       (Applicative, pure, (<*>), Alte
 import Control.Monad                             ( MonadPlus(mzero, mplus), ap, liftM, msum
                                                  )
 import Control.Monad.Base                        ( MonadBase, liftBase )
+import Control.Monad.Catch                       ( MonadCatch(..), MonadThrow(..) )
 import Control.Monad.Error                       ( ErrorT(ErrorT), runErrorT
                                                  , Error, MonadError, throwError
                                                  , catchError, mapErrorT
@@ -18,6 +19,7 @@ import Control.Monad.Reader                      ( ReaderT(ReaderT), runReaderT
 import qualified Control.Monad.RWS.Lazy as Lazy       ( RWST, mapRWST )
 import qualified Control.Monad.RWS.Strict as Strict   ( RWST, mapRWST )
 
+import Control.Monad.Trans.Except                ( ExceptT, mapExceptT )
 import Control.Monad.State.Class                      ( MonadState, get, put )
 import qualified Control.Monad.State.Lazy as Lazy     ( StateT, mapStateT )
 import qualified Control.Monad.State.Strict as Strict ( StateT, mapStateT )
@@ -58,6 +60,12 @@ type ServerPart a = ServerPartT IO a
 newtype ServerPartT m a = ServerPartT { unServerPartT :: ReaderT Request (WebT m) a }
     deriving (Monad, MonadPlus, Functor)
 
+instance MonadCatch m => MonadCatch (ServerPartT m) where
+    catch action handle = ServerPartT $ catch (unServerPartT action) (unServerPartT . handle)
+
+instance MonadThrow m => MonadThrow (ServerPartT m) where
+    throwM = ServerPartT . throwM
+
 instance MonadBase b m => MonadBase b (ServerPartT m) where
     liftBase = lift . liftBase
 
@@ -65,6 +73,19 @@ instance (MonadIO m) => MonadIO (ServerPartT m) where
     liftIO = ServerPartT . liftIO
     {-# INLINE liftIO #-}
 
+#if MIN_VERSION_monad_control(1,0,0)
+instance MonadTransControl ServerPartT where
+    type StT ServerPartT a = StT WebT (StT (ReaderT Request) a)
+    liftWith f = ServerPartT $ liftWith $ \runReader ->
+                                 liftWith $ \runWeb ->
+                                   f $ runWeb . runReader . unServerPartT
+    restoreT = ServerPartT . restoreT . restoreT
+
+instance MonadBaseControl b m => MonadBaseControl b (ServerPartT m) where
+    type StM (ServerPartT m) a = ComposeSt ServerPartT m a
+    liftBaseWith = defaultLiftBaseWith
+    restoreM     = defaultRestoreM
+#else
 instance MonadTransControl ServerPartT where
     newtype StT ServerPartT a = StSP {unStSP :: StT WebT (StT (ReaderT Request) a)}
     liftWith f = ServerPartT $ liftWith $ \runReader ->
@@ -76,6 +97,7 @@ instance MonadBaseControl b m => MonadBaseControl b (ServerPartT m) where
     newtype StM (ServerPartT m) a = StMSP {unStMSP :: ComposeSt ServerPartT m a}
     liftBaseWith = defaultLiftBaseWith StMSP
     restoreM     = defaultRestoreM     unStMSP
+#endif
 
 -- | Particularly useful when combined with 'runWebT' to produce
 -- a @m ('Maybe' 'Response')@ from a 'Request'.
@@ -251,6 +273,12 @@ filterFun = Set . Dual . Endo
 newtype FilterT a m b = FilterT { unFilterT :: Lazy.WriterT (FilterFun a) m b }
    deriving (Functor, Applicative, Monad, MonadTrans)
 
+instance MonadCatch m => MonadCatch (FilterT a m) where
+    catch action handle = FilterT $ catch (unFilterT action) (unFilterT . handle)
+
+instance MonadThrow m => MonadThrow (FilterT a m) where
+    throwM = FilterT . throwM
+
 instance MonadBase b m => MonadBase b (FilterT a m) where
     liftBase = lift . liftBase
 
@@ -258,6 +286,17 @@ instance (MonadIO m) => MonadIO (FilterT a m) where
     liftIO = FilterT . liftIO
     {-# INLINE liftIO #-}
 
+#if MIN_VERSION_monad_control(1,0,0)
+instance MonadTransControl (FilterT a) where
+    type StT (FilterT a) b = StT (Lazy.WriterT (FilterFun a)) b
+    liftWith f = FilterT $ liftWith $ \run -> f $ run . unFilterT
+    restoreT = FilterT . restoreT
+
+instance MonadBaseControl b m => MonadBaseControl b (FilterT a m) where
+    type StM (FilterT a m) c = ComposeSt (FilterT a) m c
+    liftBaseWith = defaultLiftBaseWith
+    restoreM     = defaultRestoreM
+#else
 instance MonadTransControl (FilterT a) where
     newtype StT (FilterT a) b = StFilter {unStFilter :: StT (Lazy.WriterT (FilterFun a)) b}
     liftWith f = FilterT $ liftWith $ \run -> f $ liftM StFilter . run . unFilterT
@@ -267,6 +306,7 @@ instance MonadBaseControl b m => MonadBaseControl b (FilterT a m) where
     newtype StM (FilterT a m) c = StMFilter {unStMFilter :: ComposeSt (FilterT a) m c}
     liftBaseWith = defaultLiftBaseWith StMFilter
     restoreM     = defaultRestoreM     unStMFilter
+#endif
 
 -- | A set of functions for manipulating filters.
 --
@@ -304,6 +344,12 @@ instance (Monad m) => FilterMonad a (FilterT a m) where
 newtype WebT m a = WebT { unWebT :: ErrorT Response (FilterT (Response) (MaybeT m)) a }
     deriving (Functor)
 
+instance MonadCatch m => MonadCatch (WebT m) where
+    catch action handle = WebT $ catch (unWebT action) (unWebT . handle)
+
+instance MonadThrow m => MonadThrow (WebT m) where
+    throwM = WebT . throwM
+
 instance MonadBase b m => MonadBase b (WebT m) where
     liftBase = lift . liftBase
 
@@ -311,6 +357,24 @@ instance (MonadIO m) => MonadIO (WebT m) where
     liftIO = WebT . liftIO
     {-# INLINE liftIO #-}
 
+#if MIN_VERSION_monad_control(1,0,0)
+instance MonadTransControl WebT where
+    type StT WebT a = StT MaybeT
+                       (StT (FilterT Response)
+                        (StT (ErrorT Response) a))
+    liftWith f = WebT $ liftWith $ \runError ->
+                          liftWith $ \runFilter ->
+                            liftWith $ \runMaybe ->
+                              f $ runMaybe .
+                                   runFilter .
+                                    runError . unWebT
+    restoreT = WebT . restoreT . restoreT . restoreT
+
+instance MonadBaseControl b m => MonadBaseControl b (WebT m) where
+    type StM (WebT m) a = ComposeSt WebT m a
+    liftBaseWith = defaultLiftBaseWith
+    restoreM     = defaultRestoreM
+#else
 instance MonadTransControl WebT where
     newtype StT WebT a = StWeb {unStWeb :: StT MaybeT
                                              (StT (FilterT Response)
@@ -327,7 +391,7 @@ instance MonadBaseControl b m => MonadBaseControl b (WebT m) where
     newtype StM (WebT m) a = StMWeb {unStMWeb :: ComposeSt WebT m a}
     liftBaseWith = defaultLiftBaseWith StMWeb
     restoreM     = defaultRestoreM     unStMWeb
-
+#endif
 -- | 'UnWebT' is almost exclusively used with 'mapServerPartT'. If you
 -- are not using 'mapServerPartT' then you do not need to wrap your
 -- head around this type. If you are -- the type is not as complex as
@@ -686,3 +750,22 @@ instance (Error e, FilterMonad a m) => FilterMonad a (ErrorT e m) where
 
 instance (Error e, WebMonad a m) => WebMonad a (ErrorT e m) where
     finishWith    = lift . finishWith
+
+-- ExceptT
+
+instance ServerMonad m => ServerMonad (ExceptT e m) where
+    askRq     = lift askRq
+    localRq f = mapExceptT $ localRq f
+
+instance (FilterMonad a m) => FilterMonad a (ExceptT e m) where
+    setFilter f   = lift $ setFilter f
+    composeFilter = lift . composeFilter
+    getFilter m = mapExceptT (\m' ->
+                                 do (eb, f) <- getFilter m'
+                                    case eb of
+                                      (Left e)  -> return (Left e)
+                                      (Right b) -> return $ Right (b, f)
+                  ) m
+
+instance WebMonad a m => WebMonad a (ExceptT e m) where
+    finishWith    = lift . finishWith
